/*******************************************
 *
 * Reversi 2
 *
 * Version: 1.0
 *
 * Author:  Frank Buss, fb@zuse.com
 *
 * Date:    02/21/1998
 *
 *******************************************/

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "reversi2.h"

// database info
#define Reversi2AppID    'Rev2'
#define Reversi2DBType   'Data'
char Reversi2DBName[] = "Reversi 2 game info";
VoidHand gbHandle_Reversi2;
DmOpenRef gbReversi2DB;

// board and stone size
#define BoardWidth 8
#define BoardHeight 8
#define StoneWidth 20
#define StoneHeight 18

// y-position of the board
#define BoardOffset 16

// game info structure
struct Reversi2_t {
    SByte board[BoardWidth * BoardHeight];  // the stones
    SByte color;  // the color (0 or 1) for the next stone to set
};

// local copy of the game info structure
struct Reversi2_t gbReversi2;

// the current high score
char gbScore[] = "S:   W:  ";


/*******************************************
 *
 * clear the board in the memory and set the start stones
 *
 *******************************************/
static void ClearBoard()
{
    SByte c;

    // clear the board
    for (c = 0; c < BoardWidth * BoardHeight; c++) {
        gbReversi2.board[c] = -1;
    }

    // set the start stones
    gbReversi2.board[3 + 3 * BoardWidth] = 0;
    gbReversi2.board[4 + 4 * BoardWidth] = 0;
    gbReversi2.board[4 + 3 * BoardWidth] = 1;
    gbReversi2.board[3 + 4 * BoardWidth] = 1;

    // set the color for the next stone
    gbReversi2.color = 0;
}

/*******************************************
 *
 * converts a number to a string (only 2 digits)
 *
 *******************************************/
void IToA(SByte i, char *a)
{
    if (i < 10) {
        a[0] = i + '0';
        a[1] = ' ';
    } else {
        a[0] = i / 10 + '0';
        a[1] = i - (i / 10) * 10 + '0';
    }
}


/*******************************************
 *
 * updates the score text
 *
 *******************************************/
static void UpdateScore()
{
    SByte c, b = 0, w = 0;
    RectangleType rect;

    // count the stones
    for (c = 0; c < BoardWidth * BoardHeight; c++) {
        switch (gbReversi2.board[c]) {
            case 0: b++; break;
            case 1: w++; break;
        }
    }

    // update the score string
    IToA(b, &gbScore[2]);
    IToA(w, &gbScore[7]);

    // show the score
    rect.topLeft.x = 56;
    rect.topLeft.y = 0;
    rect.extent.x = 43;
    rect.extent.y = 13;
    WinEraseRectangle(&rect, 0);
    WinDrawChars(gbScore, 9, 56, 1);
}


/*******************************************
 *
 * draws or clears a stone
 *
 *******************************************/
static DrawStone(SByte stone, SWord x, SWord y)
{
    register SWord xp = x * StoneWidth;
    register SWord yp = y * StoneHeight + BoardOffset;

    // set or clear the stone in the display    
    if (stone >= 0) {
        VoidHand bitmapHandle = DmGet1Resource('Tbmp', stone ? bitmapID_white : bitmapID_black);
        BitmapPtr bitmapP = MemHandleLock(bitmapHandle);
        WinDrawBitmap(bitmapP, xp, yp);
        MemHandleUnlock(bitmapHandle);
    } else {
        RectangleType rect;
        rect.topLeft.x = xp;
        rect.topLeft.y = yp;
        rect.extent.x = StoneWidth;
        rect.extent.y = StoneHeight;
        WinEraseRectangle(&rect, 0);
    }
}


/*******************************************
 *
 * draws the board
 *
 *******************************************/
static void ShowBoard()
{
    SWord x, y;
    for (x = 0; x < BoardWidth; x++) {
        for (y = 0; y < BoardHeight; y++)
            DrawStone(gbReversi2.board[x + y * BoardHeight], x, y);
    }
    UpdateScore();
}

    
/*******************************************
 *
 * tests a direction and sets the stones, if 'set' is true
 *
 *******************************************/
static Boolean Check(
    Boolean set,
    SWord x,
    SWord y,
    Word dir)
{
    // all 8 directions for turn stones
    const SWord directions[8][2] =
    {
        {1, 0},
        {-1, 0},
        {0, 1},
        {0, -1},
        {1, 1},
        {-1, -1},
        {-1, 1},
        {1, -1}
    };

    // get the directions
    SWord dx = directions[dir][0], dy = directions[dir][1];

    Boolean otherColor = false;
    SByte *color;

    while (true) {

        // next position
        x += dx;
        y += dy;

        // return false, if it is out of the board
        if (x > 7 || x < 0 || y > 7 || y < 0) return false;

        // get the pointer to the current field        
        color = &(gbReversi2.board[x + y * BoardHeight]);

        // return false, if it is an empty field
        if (*color < 0) return false;

        // if on the current field is a stone with the
        // current color and on a field before was another
        // color, return true
        // return false, if there was no other color before
        if (*color == gbReversi2.color) {
            if (otherColor) return true; else return false;
        } else {

            // set the otherColor-flag and draw the stone,
            // if 'set' is true
            otherColor = true;
            if (set)
                DrawStone(*color = gbReversi2.color, x, y);
        }
    }
}


/*******************************************
 *
 * sets a stone and turns other stones
 *
 *******************************************/
static void SetStone(Word x, Word y)
{
    // set a stone only if the field is empty
    if (gbReversi2.board[x + y * BoardHeight] < 0) {

        // check, if there are stones to turn
        SByte c;
        Boolean checked = false;
        for (c = 0; c < 8; c++) {
            if (Check(false, x, y, c)) {
                Check(true, x, y, c);
                checked = true;
            }
        }
        if (checked) {
            gbReversi2.board[x + y * BoardHeight] = gbReversi2.color;
            DrawStone(gbReversi2.color, x, y);
            gbReversi2.color = 1 - gbReversi2.color;
            UpdateScore();
        }
    }
}


/*******************************************
 *
 * open the database for the last board
 * and creates one, if it doesn't exist
 *
 *******************************************/
static Boolean OpenDatabase()
{
    Ptr RecPointer;
    SByte *p1, *p2;
    SByte c;
    UInt index;

    // open the database
    gbReversi2DB = DmOpenDatabaseByTypeCreator(
        Reversi2DBType,
        Reversi2AppID,
        dmModeReadWrite);

    // if there are a database, use it, otherwise create a new one,
    // open it and create a new record for the game info structure
    if (gbReversi2DB) {

        // copy the game info structure from the record
        gbHandle_Reversi2 = DmGetRecord(gbReversi2DB, 0);
        RecPointer = MemHandleLock(gbHandle_Reversi2);
        p1 = (SByte *) &gbReversi2;
        p2 = (SByte *) RecPointer;
        for (c = 0; c < sizeof(struct Reversi2_t); c++, p1++, p2++) *p1 = *p2;
        MemPtrUnlock(RecPointer);
    } else {

        // create the database
        if (DmCreateDatabase(
    		0,
	    	Reversi2DBName,
    		Reversi2AppID,
	    	Reversi2DBType,
		    false) ) return false;

        // create a record    
        gbReversi2DB = DmOpenDatabaseByTypeCreator(
            Reversi2DBType,
            Reversi2AppID,
            dmModeReadWrite);
        if (!gbReversi2DB) return false;
        gbHandle_Reversi2 = DmNewRecord(
            gbReversi2DB,
            &index,
            sizeof(struct Reversi2_t));
        ClearBoard();
    }
    return true;
}


/*******************************************
 *
 * is called on application stop
 *
 *******************************************/
static void CloseDatabase()
{
    // save the board
    Ptr RecPointer = MemHandleLock(gbHandle_Reversi2);
    DmWrite(RecPointer, 0, &gbReversi2, sizeof(struct Reversi2_t));
    MemPtrUnlock(RecPointer); 

    // release the handle and close the database
    DmReleaseRecord(gbReversi2DB, 0, true);  
    DmCloseDatabase(gbReversi2DB);
}


/*******************************************
 *
 * the event handler for the reversi main form
 *
 *******************************************/
static Boolean EventHandler(EventPtr e)
{
    Boolean   handled = false;
    SWord x, y;

    switch (e->eType) {
        case frmOpenEvent:
 
            // draw the form and show the board
            FrmDrawForm(FrmGetActiveForm());
            ShowBoard();
            handled = true;
            break;
    
        case ctlSelectEvent:
            ClearBoard();
            ShowBoard();
            handled = true;
            break;

        // menue event, used for the about dialog
        case menuEvent:
            switch (e->data.menu.itemID) {
                case menuitemID_about:
                    FrmAlert(alertID_about);
                    break;
            }
            handled = true;
            break;

        case penDownEvent:
        	if (e->screenX >= 0 && e->screenY > BoardOffset && e->screenX < 160 && e->screenY < 160) {
        	    WinWindowToDisplayPt(&e->screenX, &e->screenY);
        	    e->screenY -= BoardOffset;
        	    x = e->screenX / StoneWidth;
        	    y = e->screenY / StoneHeight;
        	    SetStone(x, y);
        	}
        	break;

    }
    return handled;
}


/*******************************************
 *
 * the main function
 *
 *******************************************/
DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
    if (cmd == sysAppLaunchCmdNormalLaunch) {
        short err;
        FormPtr form;
        EventType e;

        // open database
        if (!OpenDatabase()) return 1;

        // show the main form
        FrmGotoForm(formID_reversi2);

        // event loop
        do {
            EvtGetEvent(&e, -1);

            if (SysHandleEvent(&e)) continue;
            if (MenuHandleEvent((void *)0, &e, &err)) continue;

            if (e.eType == frmLoadEvent) {

                // init the main reversi form
                form = FrmInitForm(e.data.frmLoad.formID);
                FrmSetActiveForm(form);
                FrmSetEventHandler(form, (FormEventHandlerPtr) EventHandler);
            }
            FrmDispatchEvent(&e);
        } while(e.eType != appStopEvent);

        // close database
        CloseDatabase();
    }
    return 0;
}
